from mixcoatl.settings.load_settings import settings
import mixcoatl.auth as auth
import requests as r
from mixcoatl.decorators.lazy import lazy_property

class Resource(object):
    def __init__(self, base_path=None, request_details = 'extended'):
        if base_path is None:
            try:
                self.__path = self.__class__.path
            except:
                raise("you must override base_path")
        else:
            self.__path = base_path
        self.__last_request = None
        self.__last_error = None
        self.__current_job = None
        self.__request_details = request_details
        self.pending_changes = {}

    def __props(self):
        p = [k for k,v in self.__class__.__dict__.items() if type(v) is lazy_property]
        rp = ['last_error', 'path', 'last_request', 'current_job']
        return p + rp


    def __repr__(self):
        from mixcoatl.utils import convert
        d = {}
        for x in self.__props():
            try:
                if x == 'last_request':
                    d[x] = str(getattr(self, x))
                else:
                    d[x] = getattr(self, x)
            except AttributeError:
                d[x] = None
            except KeyError:
                d[x] = None
        return repr(convert(d))

    @property
    def request_details(self):
        return self.__request_details

    @request_details.setter
    def request_details(self, level):
        self.__request_details = level

    @property
    def path(self):
        return self.__path

    @path.setter
    def path(self, p):
        self.__path = p

    @property
    def last_request(self):
        return self.__last_request

    @last_request.setter
    def last_request(self, lr):
        self.__last_request = lr

    @property
    def last_error(self):
        return self.__last_error

    @last_error.setter
    def last_error(self, le):
        self.__last_error = le

    @property
    def current_job(self):
        return self.__current_job

    @current_job.setter
    def current_job(self, cj):
        self.__current_job = cj

    def load(self):
        from mixcoatl.utils import uncamel_keys
        p = self.path+"/"+str(getattr(self, self.__class__.primary_key))

        #self.request_details = 'extended'
        s = self.get(p)
        if self.last_error is None:
            try:
                scope = uncamel_keys(s[self.__class__.collection_name][0])
                for k in scope.keys():
                    nk = '_%s__%s' % (self.__class__.__name__, k)
                    setattr(self, nk, scope[k])
                self.loaded = True
            except AttributeError:
                print("missing attribute: "+k)
        else:
            return self.last_error

    def __doreq(self, method, *args, **kwargs):
        failures = [400, 403, 404, 409, 500, 501, 503]
        sig = auth.get_sig(method, self.path)
        url = settings.endpoint+'/'+self.path

        headers = {'x-esauth-access': sig['access_key'],
        'x-esauth-timestamp': str(sig['timestamp']),
        'x-esauth-signature': str(sig['signature']),
        'x-es-details': self.__request_details,
        'Accept': 'application/json',
        'User-Agent': sig['ua']}

        results = getattr(r, method.lower())(url, headers=headers, *args, **kwargs)

        self.last_error = None
        self.last_request = results
        if results.status_code in failures:
            self.last_error = results.json()
            return self.last_error

        if method == 'GET':
            try:
                results.raise_for_status()
                self.last_error = None
                return results.json()
            except r.exceptions.HTTPError:
                self.last_error = results.json()
                return False
        if method == 'DELETE':
            if results.status_code !=204:
                self.last_error = results.json()
                return False
            else:
                return True
        if method == 'PUT':
            if results.status_code == 202:
                self.current_job = results.json()['jobs'][0]['jobId']
                return results.json()
            elif results.status_code == 204:
                return True
            else:
                self.last_error = results.json()
                return self.last_error
        if method == 'POST':
            if results.status_code in [201,202]:
                if results.status_code == 202:
                    self.current_job = results.json()['jobs'][0]['jobId']
                return results.json()
            else:
                self.last_error = results.json()
                return False

    def set_path(self, path=None):
        if path is None:
            path = self.path
        else:
            self.path = path

    def get(self, path=None, *args, **kwargs):
        self.set_path(path)
        return self.__doreq('GET', *args, **kwargs)

    def post(self, path=None, *args, **kwargs):
        self.set_path(path)
        return self.__doreq('POST', *args, **kwargs)

    def put(self, path=None, *args, **kwargs):
        self.set_path(path)
        return self.__doreq('PUT', *args, **kwargs)

    def delete(self, path=None, *args, **kwargs):
        self.set_path(path)
        return self.__doreq('DELETE', *args, **kwargs)

    def pprint(self):
        import pprint
        pprint.pprint(eval(repr(self)))

    def to_dict(self):
        return eval(repr(self))

    def track_change(self, var, new_value):
        prev = getattr(self, var)
        self.pending_changes[var] = {'old': prev, 'new':new_value}



import time

# TODO: refactor this out as it creates a circular import issue
def wait_for_job(job_id, status='COMPLETE'):
    from mixcoatl.admin import job
    j = job.get(job_id)
    initial_status = j['status']
    if initial_status == 'ERROR':
        return False
    while job.get(job_id)['status'] != status:
        time.sleep(5)
        j = job.get(job_id)
        if j['status'] == 'ERROR':
            return False
        else:
            continue

    return True

def uncamel(str):
    import re
    s = lambda str: re.sub('(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', str).lower().strip('_')
    return s(str)

def uncamel_keys(d1):
    d2 = dict()
    if not isinstance(d1, dict):
        return d1
    for k,v in d1.iteritems():
        new_key = uncamel(k)
        if isinstance(v, dict):
            d2[new_key] = uncamel_keys(v)
        elif isinstance(v, list):
            d2[new_key] = [uncamel_keys(item) for item in v]
        else:
            d2[new_key] = v
    return d2

def camelize(str):
    s = ''.join([t.title() for t in str.split('_')])
    return s[0].lower()+s[1:]

def camel_keys(d1):
    d2 = dict()
    if not isinstance(d1, dict):
        return d1
    for k, v in d1.iteritems():
        new_key = camelize(k)
        if isinstance(v, dict):
            d2[new_key] = camel_keys(v)
        elif isinstance(v, list):
            d2[new_key] = [camel_keys(item) for item in v]
        else:
            d2[new_key] = v
    return d2

def convert(input):
    if isinstance(input, dict):
        return dict((convert(key), convert(value)) for key, value in input.iteritems())
    elif isinstance(input, list):
        return [convert(element) for element in input]
    elif isinstance(input, unicode):
        return input.encode('utf-8')
    else:
        return input


class ValidationException(Exception): pass

def required_attrs(attrs):
    def wrapper(method):
        def validate(obj, *args, **kwargs):
            for ra in attrs:
                try:
                    if getattr(obj, ra) is None:
                        raise ValidationException('Required attribute "%s" is missing' % ra)
                except AttributeError:
                    raise ValidationException('Required attribute "%s" is missing' % ra)
            return method(obj, *args, **kwargs)
        return validate
    return wrapper




from mixcoatl.resource import Resource
from mixcoatl.admin import job
from mixcoatl.utils import wait_for_job
from mixcoatl.utils import camel_keys
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.decorators.lazy import lazy_property

import json


class Server(Resource):
    path = 'infrastructure/Server'
    collection_name = 'servers'
    primary_key = 'server_id'

    def __init__(self, server_id = None, *args, **kwargs):
        self.collection_name = self.__class__.collection_name
        Resource.__init__(self)
        self.__server_id = server_id

    @property
    def server_id(self):
        '''The immutable enStratus ID of this server'''
        return self.__server_id

    @lazy_property
    def agent_version(self):
        return self.__agent_version

    @lazy_property
    def cloud(self):
        '''The id of the cloud where the instance is located'''
        return self.__cloud

    @lazy_property
    def label(self):
        return self.__label

    @label.setter
    def label(self, l):
        self.__label = l

    @lazy_property
    def customer(self):
        return self.__customer

    @lazy_property
    def data_center(self):
        '''The id of the specific datacenter where the instance is located.'''
        return self.__data_center

    @data_center.setter
    def data_center(self, d):
        self.__data_center = {u'data_center_id':d}

    @lazy_property
    def description(self):
        return self.__description

    @description.setter
    def description(self, d):
        self.__description = d

    @lazy_property
    def machine_image(self):
        return self.__machine_image

    @machine_image.setter
    def machine_image(self, m):
        self.__machine_image = {u'machine_image_id':m}

    @lazy_property
    def firewalls(self):
        return self.__firewalls

    @firewalls.setter
    def firewalls(self, f):
        self.__firewalls = f

    @lazy_property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n

    @lazy_property
    def owning_groups(self):
        return self.__owning_groups

    @lazy_property
    def owning_user(self):
        return self.__owning_user

    @lazy_property
    def platform(self):
        return self.__platform

    @lazy_property
    def private_ip_addresses(self):
        return self.__private_ip_addresses

    @lazy_property
    def public_ip_address(self):
        return self.__public_ip_address

    @lazy_property
    def region(self):
        return self.__region

    @lazy_property
    def platform(self):
        return self.__platform

    @lazy_property
    def provider_product_id(self):
        return self.__provider_product_id

    @provider_product_id.setter
    def provider_product_id(self, id):
        self.__provider_product_id = id

    @lazy_property
    def provider_id(self):
        return self.__provider_id

    @lazy_property
    def start_date(self):
        return self.__start_date

    @lazy_property
    def status(self):
        return self.__status

    @lazy_property
    def budget(self):
        return self.__budget

    @property
    def keypair(self):
        return self.__keypair

    @keypair.setter
    def keypair(self, kp):
        self.__keypair = kp

    def reload(self):
        if self.server_id is not None:
            self.load()
        elif self.current_job is None:
            self.load()
        else:
            if wait_for_job(self.current_job):
                self.server_id = job.get(self.current_job)['message']
                self.load()
            else:
                return self.last_error

    @required_attrs(['server_id'])
    def destroy(self, reason='no reason provided'):
        p = self.path+"/"+str(self.server_id)
        qopts = {'reason':reason}
        return self.delete(p, params=qopts)

    @required_attrs(['server_id'])
    def pause(self, reason=None):
        p = '%s/%s' % (self.path, str(self.server_id))
        payload = {'pause':[{}]}

        if reason is not None:
            payload['pause'][0].update({'reason':reason})

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['provider_product_id', 'machine_image', 'description', 'name','data_center'])
    def launch(self, callback=None):
        optional_attrs = ['firewalls','keypair', 'label']
        if self.server_id is not None:
            raise ServerLaunchException('Cannot launch an already running server: %s' % self.server_id)

        payload = {'launch':
                    [{
                        'productId': self.provider_product_id,
                        'machineImage': camel_keys(self.machine_image),
                        'description': self.description,
                        'name': self.name,
                        'dataCenter': camel_keys(self.data_center)
                    }]}

        for oa in optional_attrs:
            try:
                if getattr(self, oa) is not None:
                    payload['launch'][0].update(camel_keys({oa:getattr(self, oa)}))
            except AttributeError:
                pass

        s = self.post(data=json.dumps(payload))
        if callback is not None:
            callback(s)
        else:
            return s

    def duplicate(self, server):
        pass

    @classmethod
    def all(cls):
        r = Resource(cls.path)
        s = r.get()
        if r.last_error is None:
            return [cls(server['serverId']) for server in s[cls.collection_name]]
        else:
            return r.last_error

class ServerException(BaseException): pass
class ServerLaunchException(ServerException): pass



from mixcoatl.resource import Resource

class Job(Resource):
    path = 'admin/Job'
    collection_name = 'jobs'
    def __init__(self):
        Resource.__init__(self)

    def get_job(self, job_id):
        p = self.path+"/"+str(job_id)
        return self.get(path=p)

    @classmethod
    def all(cls):
        r = Resource(cls.path)
        jobs = r.get()[cls.collection_name]
        if r.last_error is None:
            return jobs
        else:
            return r.last_error

def get(job_id):
    j = Job()
    job = j.get_job(job_id)
    if j.last_error is None:
        return job['jobs'][0]
    else:
        return j.last_error
